from typing import Iterable, Iterator, TypeVar

T = TypeVar("T")


def merge_sorted(its: Iterable[Iterable[T]]) -> Iterator[T]:
    its = [iter(it) for it in its]
    sentinel = object()
    items = [None for _ in its]
    while its:
        to_del = set()
        pick = None
        for i, it in enumerate(its):
            if items[i] is None:
                item = next(it, sentinel)
                if item is sentinel:
                    to_del.add(i)
                    continue
                else:
                    items[i] = item
            if pick is None or items[i] < items[pick]:
                pick = i
        if pick is None:
            return
        yield items[pick]
        items[pick] = None
        for i in to_del:
            del items[i]
            del its[i]



import bisect
import calendar
import os
from datetime import date
from typing import List, Optional, Tuple, Type

import pytest
import yaml

from earthkit.time.calendar import (
    FRIDAY,
    MONDAY,
    SATURDAY,
    SUNDAY,
    THURSDAY,
    WEDNESDAY,
    month_length,
)
from earthkit.time.sequence import (
    DailySequence,
    MonthlySequence,
    Sequence,
    WeeklySequence,
    YearlySequence,
)


class ExcludeLeapFeb28:
    def __contains__(self, day: date) -> bool:
        if not day.month == 2:
            return False
        if not calendar.isleap(day.year):
            return False
        return day.day == 28


class ExcludeNonTenYears:
    def __contains__(self, day: date) -> bool:
        return day.year % 10 != 0


@pytest.mark.parametrize(
    "seq, ymds, outside",
    [
        pytest.param(
            DailySequence(),
            [(1983, 4, 28), (1983, 4, 29), (1983, 4, 30), (1983, 5, 1), (1983, 5, 2)],
            [],
            id="daily-simple",
        ),
        pytest.param(
            DailySequence(),
            [
                (2001, 12, 29),
                (2001, 12, 30),
                (2001, 12, 31),
                (2002, 1, 1),
                (2002, 1, 2),
            ],
            [],
            id="daily-crossyear",
        ),
        pytest.param(
            DailySequence(),
            [(2003, 2, 26), (2003, 2, 27), (2003, 2, 28), (2003, 3, 1), (2003, 3, 2)],
            [],
            id="daily-feb28-nonleap",
        ),
        pytest.param(
            DailySequence(),
            [(2016, 2, 27), (2016, 2, 28), (2016, 2, 29), (2016, 3, 1), (2016, 3, 2)],
            [],
            id="daily-feb28-leap",
        ),
        pytest.param(
            DailySequence(excludes=[1]),
            [(1995, 4, 28), (1995, 4, 29), (1995, 4, 30), (1995, 5, 2), (1995, 5, 3)],
            [(1995, 5, 1)],
            id="daily-exclude",
        ),
        pytest.param(
            DailySequence(excludes=[29]),
            [(2000, 2, 26), (2000, 2, 27), (2000, 2, 28), (2000, 3, 1), (2000, 3, 2)],
            [(2000, 2, 29)],
            id="daily-exclude-leap",
        ),
        pytest.param(
            DailySequence(excludes=range(1, 29)),
            [(2020, 1, 30), (2020, 1, 31), (2020, 2, 29), (2020, 3, 29), (2020, 3, 30)],
            [(2020, 2, 20)],
            id="daily-exclude-almostall",
        ),
        pytest.param(
            WeeklySequence(2),
            [(1999, 3, 24), (1999, 3, 31), (1999, 4, 7), (1999, 4, 14), (1999, 4, 21)],
            [(1999, 4, 20)],
            id="weekly-simple",
        ),
        pytest.param(
            WeeklySequence(FRIDAY),
            [
                (2011, 12, 23),
                (2011, 12, 30),
                (2012, 1, 6),
                (2012, 1, 13),
                (2012, 1, 20),
            ],
            [(2012, 1, 2)],
            id="weekly-crossyear",
        ),
        pytest.param(
            WeeklySequence([2, 5]),
            [(2007, 2, 24), (2007, 2, 28), (2007, 3, 3), (2007, 3, 7), (2007, 3, 10)],
            [(2007, 2, 25), (2007, 3, 5)],
            id="weekly-feb28-nonleap",
        ),
        pytest.param(
            WeeklySequence([MONDAY, THURSDAY]),
            [(2024, 2, 22), (2024, 2, 26), (2024, 2, 29), (2024, 3, 4), (2024, 3, 7)],
            [(2024, 2, 28), (2024, 3, 2)],
            id="weekly-feb28-leap",
        ),
        pytest.param(
            MonthlySequence(15),
            [(1989, 3, 15), (1989, 4, 15), (1989, 5, 15), (1989, 6, 15), (1989, 7, 15)],
            [(1989, 4, 30), (1989, 5, 19)],
            id="monthly-simple",
        ),
        pytest.param(
            MonthlySequence([7, 21]),
            [
                (2014, 11, 21),
                (2014, 12, 7),
                (2014, 12, 21),
                (2015, 1, 7),
                (2015, 1, 21),
            ],
            [(2014, 12, 14), (2014, 12, 31)],
            id="monthly-crossyear",
        ),
        pytest.param(
            MonthlySequence(range(1, 32, 7)),
            [(2009, 2, 15), (2009, 2, 22), (2009, 3, 1), (2009, 3, 8), (2009, 3, 15)],
            [(2009, 3, 14)],
            id="monthly-feb28-nonleap",
        ),
        pytest.param(
            MonthlySequence([28, 29]),
            [(1992, 1, 29), (1992, 2, 28), (1992, 2, 29), (1992, 3, 28), (1992, 3, 29)],
            [(1992, 2, 18), (1992, 3, 14)],
            id="monthly-feb28-leap",
        ),
        pytest.param(
            MonthlySequence([11, 22], excludes=[(11, 11)]),
            [
                (1987, 10, 11),
                (1987, 10, 22),
                (1987, 11, 22),
                (1987, 12, 11),
                (1987, 12, 22),
            ],
            [(1987, 11, 11)],
            id="monthly-exclude",
        ),
        pytest.param(
            MonthlySequence([27, 29, 31], excludes=[(2, 29)]),
            [(2008, 1, 31), (2008, 2, 27), (2008, 3, 27), (2008, 3, 29), (2008, 3, 31)],
            [(2008, 2, 29), (2008, 3, 28)],
            id="monthly-exclude-leap",
        ),
        pytest.param(
            MonthlySequence(31, excludes=[(i, 31) for i in range(1, 10)]),
            [
                (2021, 10, 31),
                (2021, 12, 31),
                (2022, 10, 31),
                (2022, 12, 31),
                (2023, 10, 31),
            ],
            [(2022, 5, 9), (2022, 6, 1)],
            id="monthly-exclude-almostall",
        ),
        pytest.param(
            YearlySequence((1, 1)),
            [(1999, 1, 1), (2000, 1, 1), (2001, 1, 1), (2002, 1, 1), (2003, 1, 1)],
            [(2000, 7, 2), (2002, 2, 2)],
            id="yearly-simple",
        ),
        pytest.param(
            YearlySequence([(1, 1), (4, 2), (7, 2), (10, 1)]),
            [
                (2017, 7, 2),
                (2017, 10, 1),
                (2018, 1, 1),
                (2018, 4, 2),
                (2018, 7, 2),
            ],
            [(2017, 11, 16), (2018, 6, 18)],
            id="yearly-crossyear",
        ),
        pytest.param(
            YearlySequence([(2, 28), (2, 29), (3, 1)]),
            [(1994, 2, 28), (1994, 3, 1), (1995, 2, 28), (1995, 3, 1), (1996, 2, 28)],
            [(1994, 8, 30), (1995, 2, 22)],
            id="yearly-feb28-nonleap",
        ),
        pytest.param(
            YearlySequence([(i, 29) for i in range(1, 13)]),
            [(2008, 1, 29), (2008, 2, 29), (2008, 3, 29), (2008, 4, 29), (2008, 5, 29)],
            [(2008, 2, 28), (2008, 5, 14)],
            id="yearly-feb28-leap",
        ),
        pytest.param(
            YearlySequence((2, 29)),
            [(2000, 2, 29), (2004, 2, 29), (2008, 2, 29), (2012, 2, 29), (2016, 2, 29)],
            [(2003, 2, 28)],
            id="yearly-leaponly",
        ),
        pytest.param(
            YearlySequence([(7, 13)], excludes=[date(2023, 7, 13)]),
            [(2020, 7, 13), (2021, 7, 13), (2022, 7, 13), (2024, 7, 13), (2025, 7, 13)],
            [(2023, 7, 13)],
            id="yearly-exclude",
        ),
        pytest.param(
            YearlySequence([(1, 31), (2, 28), (2, 29)], excludes=ExcludeLeapFeb28()),
            [(2007, 1, 31), (2007, 2, 28), (2008, 1, 31), (2008, 2, 29), (2009, 1, 31)],
            [(2007, 2, 14), (2008, 3, 10)],
            id="yearly-exclude-leap",
        ),
        pytest.param(
            YearlySequence([(3, 31)], excludes=ExcludeNonTenYears()),  # FIXME
            [(1990, 3, 31), (2000, 3, 31), (2010, 3, 31), (2020, 3, 31), (2030, 3, 31)],
            [(2001, 3, 31), (2005, 3, 31)],
            id="yearly-exclude-almostall",
        ),
    ],
)
def test_sequence(
    seq: Sequence,
    ymds: List[Tuple[int, int, int]],
    outside: List[Tuple[int, int, int]],
):
    dates = [date(y, m, d) for y, m, d in ymds]

    for d in dates:
        assert d in seq

    for i in range(1, len(dates) - 1):
        prev = dates[i - 1]
        cur = dates[i]
        next = dates[i + 1]

        assert seq.next(cur) == next
        assert seq.next(cur, False) == cur
        assert seq.previous(cur) == prev
        assert seq.previous(cur, False) == cur
        assert seq.nearest(cur) == cur
        assert seq.nearest(cur, resolve="previous") == cur
        assert seq.nearest(cur, resolve="next") == cur

    assert list(seq.range(dates[0], dates[-1])) == dates
    assert list(seq.range(dates[0], dates[-1], include_start=False)) == dates[1:]
    assert (
        list(seq.range(dates[0], dates[-1], include_start=False, include_end=False))
        == dates[1:-1]
    )
    assert list(seq.range(dates[0], dates[-1], include_end=False)) == dates[:-1]

    assert list(seq.bracket(dates[2])) == [dates[1], dates[3]]
    assert list(seq.bracket(dates[2], 2)) == [dates[0], dates[1], dates[3], dates[4]]
    assert list(seq.bracket(dates[2], (1, 2))) == [dates[1], dates[3], dates[4]]
    assert list(seq.bracket(dates[2], 1, strict=False)) == dates[1:4]
    assert list(seq.bracket(dates[2], (2, 1), strict=False)) == dates[:4]

    for out_tup in outside:
        out_date = date(*out_tup)
        out_i = bisect.bisect_left(dates, out_date)
        before = out_i
        after = len(dates) - out_i

        assert out_date not in seq
        assert seq.next(out_date) == dates[out_i]
        assert seq.previous(out_date) == dates[out_i - 1]

        db = (out_date - dates[out_i - 1]).days
        da = (dates[out_i] - out_date).days
        if db != da:
            exp_nearest = dates[out_i - 1] if db < da else dates[out_i]
            assert seq.nearest(out_date) == exp_nearest
        else:
            assert seq.nearest(out_date) == dates[out_i - 1]
            assert seq.nearest(out_date, resolve="previous") == dates[out_i - 1]
            assert seq.nearest(out_date, resolve="next") == dates[out_i]

        assert list(seq.range(out_date, dates[-1])) == dates[out_i:]
        assert (
            list(seq.range(out_date, dates[-1], include_start=False)) == dates[out_i:]
        )
        assert list(seq.range(dates[0], out_date)) == dates[:out_i]
        assert list(seq.range(dates[0], out_date, include_end=False)) == dates[:out_i]

        assert list(seq.bracket(out_date)) == [dates[out_i - 1], dates[out_i]]
        assert list(seq.bracket(out_date, (before, after))) == dates
        assert (
            list(seq.bracket(out_date, (min(2, before), after)))
            == dates[max(0, out_i - 2) :]
        )
        assert (
            list(seq.bracket(out_date, (before, min(2, after))))
            == dates[: min(len(dates), out_i + 2)]
        )


@pytest.mark.parametrize(
    "seq_dict, expect_type, expect_days, expect_excludes",
    [
        pytest.param({"type": "daily"}, DailySequence, None, set(), id="daily"),
        pytest.param(
            {"type": "daily", "excludes": [2, 4, 6]},
            DailySequence,
            None,
            {2, 4, 6},
            id="daily-excludes",
        ),
        pytest.param(
            {"type": "weekly", "days": 2},
            WeeklySequence,
            [WEDNESDAY],
            None,
            id="weekly",
        ),
        pytest.param(
            {"type": "weekly", "days": "Th"},
            WeeklySequence,
            [THURSDAY],
            None,
            id="weekly-str",
        ),
        pytest.param(
            {"type": "weekly", "days": [4, 6]},
            WeeklySequence,
            [FRIDAY, SUNDAY],
            None,
            id="weekly-intlist",
        ),
        pytest.param(
            {"type": "weekly", "days": ["mon", "sat"]},
            WeeklySequence,
            [MONDAY, SATURDAY],
            None,
            id="weekly-strlist",
        ),
        pytest.param(
            {"type": "monthly", "days": 13}, MonthlySequence, [13], set(), id="monthly"
        ),
        pytest.param(
            {"type": "monthly", "days": [7, 21]},
            MonthlySequence,
            [7, 21],
            set(),
            id="monthly-list",
        ),
        pytest.param(
            {"type": "monthly", "days": [4, 8, 12, 29], "excludes": [[2, 29]]},
            MonthlySequence,
            [4, 8, 12, 29],
            {(2, 29)},
            id="monthly-list-exclude",
        ),
        pytest.param(
            {"type": "monthly", "days": [4, 8, 12, 29], "excludes": ["0229", "0312"]},
            MonthlySequence,
            [4, 8, 12, 29],
            {(2, 29), (3, 12)},
            id="monthly-list-strexclude",
        ),
        pytest.param(
            {"type": "yearly", "days": [3, 14]},
            YearlySequence,
            [(3, 14)],
            set(),
            id="yearly",
        ),
        pytest.param(
            {"type": "yearly", "days": "1203"},
            YearlySequence,
            [(12, 3)],
            set(),
            id="yearly-str",
        ),
        pytest.param(
            {"type": "yearly", "days": [[5, 22], [8, 24]]},
            YearlySequence,
            [(5, 22), (8, 24)],
            set(),
            id="yearly-mdlist",
        ),
        pytest.param(
            {"type": "yearly", "days": ["1001", "1111"]},
            YearlySequence,
            [(10, 1), (11, 11)],
            set(),
            id="yearly-strlist",
        ),
        pytest.param(
            {
                "type": "yearly",
                "days": ["0311", "0722", "1224"],
                "excludes": [[2000, 12, 24]],
            },
            YearlySequence,
            [(3, 11), (7, 22), (12, 24)],
            {date(2000, 12, 24)},
            id="yearly-strlist-exclude",
        ),
        pytest.param(
            {
                "type": "yearly",
                "days": ["0311", "0722", "1224"],
                "excludes": ["20110311", "20070722"],
            },
            YearlySequence,
            [(3, 11), (7, 22), (12, 24)],
            {date(2011, 3, 11), date(2007, 7, 22)},
            id="yearly-strlist-strexclude",
        ),
    ],
)
def test_create_sequence(
    seq_dict: dict,
    expect_type: Type[Sequence],
    expect_days: Optional[list],
    expect_excludes: Optional[set],
):
    seq = Sequence.from_dict(seq_dict)
    assert type(seq) is expect_type
    if expect_days is not None:
        assert seq.days == expect_days
    if expect_excludes is not None:
        assert seq.excludes == expect_excludes


@pytest.mark.parametrize(
    "seq_dict, expect_msg",
    [
        pytest.param({}, "^Sequence dictionary must contain `type` key$", id="notype"),
        pytest.param(
            {"type": "sesquiannual"}, "^Unknown type 'sesquiannual'$", id="unknowntype"
        ),
        pytest.param(
            {"type": "weekly"},
            "^Weekly sequence must provide `days`$",
            id="weekly-nodays",
        ),
        pytest.param(
            {"type": "monthly"},
            "^Monthly sequence must provide `days`$",
            id="monthly-nodays",
        ),
        pytest.param(
            {"type": "yearly"},
            "^Yearly sequence must provide `days`$",
            id="yearly-nodays",
        ),
    ],
)
def test_create_sequence_invalid(seq_dict: dict, expect_msg: str):
    with pytest.raises(ValueError, match=expect_msg):
        Sequence.from_dict(seq_dict)


def test_sequence_from_resource(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
):
    seq = Sequence.from_resource("ecmwf-mon-thu")
    assert type(seq) is WeeklySequence
    assert seq.days == [MONDAY, THURSDAY]

    with pytest.raises(FileNotFoundError):
        Sequence.from_resource("invalid-sequence")

    seq_path1 = tmp_path_factory.mktemp("seqs1")
    (seq_path1 / "wednesdays.yaml").write_text(
        yaml.safe_dump({"type": "weekly", "days": ["Wednesday"]})
    )
    (seq_path1 / "foo.yaml").write_text(
        yaml.safe_dump({"type": "monthly", "days": [2, 4, 6, 8]})
    )

    seq_path2 = tmp_path_factory.mktemp("seqs2")
    (seq_path2 / "foo.yaml").write_text(yaml.safe_dump({"type": "daily"}))
    (seq_path2 / "ecmwf-4days.yaml").write_text(
        yaml.safe_dump(
            {
                "type": "yearly",
                "days": [
                    (m, d)
                    for m in range(1, 13)
                    for d in range(1, month_length(1999, m) + 1, 4)
                ],
            }
        )
    )

    envname = "EARTHKIT_TIME_SEQ_PATH"
    monkeypatch.setenv(envname, os.pathsep.join([str(seq_path1), str(seq_path2)]))

    print(os.getenv(envname))

    seq = Sequence.from_resource("wednesdays")
    assert type(seq) is WeeklySequence
    assert seq.days == [WEDNESDAY]

    seq = Sequence.from_resource("foo")
    assert type(seq) is MonthlySequence
    assert seq.days == [2, 4, 6, 8]

    seq = Sequence.from_resource("ecmwf-4days")
    assert type(seq) is YearlySequence



import calendar
from datetime import date, datetime
from enum import IntEnum
from typing import Tuple, Union


class Weekday(IntEnum):
    """:class:`enum.IntEnum` representing week days"""

    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6


MONDAY = Weekday.MONDAY
TUESDAY = Weekday.TUESDAY
WEDNESDAY = Weekday.WEDNESDAY
THURSDAY = Weekday.THURSDAY
FRIDAY = Weekday.FRIDAY
SATURDAY = Weekday.SATURDAY
SUNDAY = Weekday.SUNDAY


def to_weekday(arg: Union[int, str]) -> Weekday:
    """Convert integers and strings to weekdays

    Any unambiguous prefix of a weekday name will be accepted, and case is
    ignored.
    """
    if isinstance(arg, str) and arg.isdigit():
        arg = int(arg)
    if isinstance(arg, int):
        if arg not in range(7):
            raise ValueError(f"Week day out of range: {arg} not in 0-6")
        return Weekday(arg)
    arg = arg.upper()
    matching = [wd for wd in Weekday if wd.name.startswith(arg)]
    if not matching:
        raise ValueError(f"Unrecognised week day: {arg!r}")
    if len(matching) > 1:
        others = ", ".join(wd.name.capitalize() for wd in matching)
        raise ValueError(f"Ambiguous week day: {arg!r} could be any of {others}")
    return matching[0]


_MONTH_LENGTHS = [
    None,
    31,
    (lambda year: 29 if calendar.isleap(year) else 28),
    31,
    30,
    31,
    30,
    31,
    31,
    30,
    31,
    30,
    31,
]


def month_length(year: int, month: int) -> int:
    """Return the number of days of a given month"""
    if month < 1 or month > 12:
        raise ValueError(f"Invalid month: {month}")
    mlen = _MONTH_LENGTHS[month]
    if isinstance(mlen, int):
        return mlen
    return mlen(year)


def day_exists(year: int, month: int, day: int) -> bool:
    """Check whether a given day exists in the calendar"""
    if month < 1 or month > 12:
        return False
    if day < 1 or day > month_length(year, month):
        return False
    return True


class MonthInYear:
    """Represent a given month in a year"""

    year: int
    month: int

    def __init__(self, year: int, month: int):
        self.year = year
        if month < 1 or month > 12:
            raise ValueError(f"Invalid month: {month}")
        self.month = month

    def __contains__(self, day: Union[int, date]) -> bool:
        if isinstance(day, date):
            if day.year != self.year:
                return False
            if day.month != self.month:
                return False
            day = day.day
        if day < 1 or day > self.length():
            return False
        return True

    def length(self) -> int:
        """Returns the number of days in the given month"""
        return month_length(self.year, self.month)

    def next(self) -> "MonthInYear":
        """Return the following month"""
        d, m = divmod(self.month, 12)
        m += 1
        return MonthInYear(self.year + d, m)

    def previous(self) -> "MonthInYear":
        """Return the previous month"""
        d, m = divmod(self.month - 2, 12)
        m += 1
        return MonthInYear(self.year + d, m)


def parse_mmdd(arg: Union[Tuple[int, int], str]) -> Tuple[int, int]:
    """Convert pairs of ints or MMDD strings into (month, day) pairs"""
    if not isinstance(arg, str):
        m, d = arg
        if not day_exists(2000, m, d):
            raise ValueError(f"Invalid day: {d} not in 1-{month_length(2000, m)}")
        return (m, d)
    if len(arg) != 4:
        raise ValueError(f"Unrecognised month-day value: {arg!r}")
    mm = arg[:2]
    dd = arg[2:]
    if not mm.isdigit() or not dd.isdigit():
        raise ValueError(f"Unrecognised month-day value: {arg!r}")
    m = int(mm)
    if m not in range(1, 13):
        raise ValueError(f"Invalid month: {m} not in 1-12")
    d = int(dd)
    if not day_exists(2000, m, d):
        raise ValueError(f"Invalid day: {d} not in 1-{month_length(2000, m)}")
    return (m, d)


def parse_date(arg: Union[str, Tuple[int, int, int]]) -> date:
    """Convert triples of ints or YYYYMMDD strings into date objects"""
    if not isinstance(arg, str):
        y, m, d = arg
        if not day_exists(y, m, d):
            raise ValueError(f"Invalid date: {arg!r}")
        return date(y, m, d)
    try_formats = ["%Y%m%d"]
    dt = None
    for fmt in try_formats:
        try:
            dt = datetime.strptime(arg, fmt)
        except ValueError:
            continue
        else:
            break
    if dt is None:
        raise ValueError(f"Unrecognised date format: {arg!r}")
    return dt.date()



import argparse
from typing import Callable


class ActionParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.set_defaults(action=None)
        self.action_subparsers = self.add_subparsers(
            title="actions",
            metavar="<action>",
            help="use `%(prog)s %(metavar)s --help` for details",
            parser_class=argparse.ArgumentParser,  # otherwise we add actions recursively
        )

    def add_action(
        self,
        name: str,
        func: Callable[[argparse.ArgumentParser, argparse.Namespace], None],
        **kwargs,
    ) -> argparse.ArgumentParser:
        action = self.action_subparsers.add_parser(name, **kwargs)
        action.set_defaults(action=func)
        return action



import argparse
import re
from typing import List, Tuple

from ..calendar import Weekday, parse_date, parse_mmdd, to_weekday
from ..sequence import (
    DailySequence,
    MonthlySequence,
    Sequence,
    WeeklySequence,
    YearlySequence,
)


def weekly_days(arg: str) -> List[Weekday]:
    if not arg:
        return []
    return [to_weekday(elem) for elem in arg.split("/")]


def int_list(arg: str) -> List[int]:
    if not arg:
        return []
    return [int(elem) for elem in arg.split("/")]


def yearly_days(arg: str) -> List[Tuple[int, int]]:
    if not arg:
        return []
    return [parse_mmdd(elem) for elem in arg.split("/")]


def list_arg(arg: str) -> List[str]:
    if not arg:
        return []
    return arg.split("/")


SEQ_EPILOG = """
WEEKDAYS: week days can be specified either by number (0 = Monday, 1 = Tuesday,
etc) or by any unambiguous prefix of the name (case-insensitive, e.g. M, tue,
Friday)

PRESETS: sequence presets can be stored in the package as well as externally
defined. If a preset name is given, the corresponding file will be searched in
``EARTHKIT_TIME_SEQ_PATH``, then in the package itself

EXCLUDES: specific days can be excluded from sequences:
* daily: exclude specific days of the month
* monthly: exclude specific dates in the year (MMDD)
* yearly: exclude specific dates (YYYYMMDD)
"""


def add_sequence_args(parser: argparse.ArgumentParser):
    seq_group = parser.add_mutually_exclusive_group(required=True)
    seq_group.add_argument("--daily", action="store_true", help="daily inputs")
    seq_group.add_argument(
        "--weekly",
        type=weekly_days,
        default=None,
        help="weekly inputs on these days (see WEEKDAYS, slash-separated)",
    )
    seq_group.add_argument(
        "--monthly",
        type=int_list,
        default=None,
        help="monthly inputs on these days (slash-separated)",
    )
    seq_group.add_argument(
        "--yearly",
        type=yearly_days,
        default=None,
        help="yearly inputs on these days (MMDD, slash-separated)",
    )
    seq_group.add_argument(
        "--preset",
        default=None,
        help="name of a preset sequence, or path to a valid YAML preset file (see PRESETS)",
    )

    parser.add_argument(
        "--exclude",
        type=list_arg,
        default=[],
        help="exclude these dates (see EXCLUDES, slash-separated)",
    )


def create_sequence(
    parser: argparse.ArgumentParser, args: argparse.Namespace
) -> Sequence:
    seq = None
    try:
        if args.daily:
            try:
                excludes = [int(elem) for elem in args.exclude]
            except ValueError as e:
                raise ValueError(
                    "Invalid excludes, must be a slash-separated list of days"
                ) from e
            seq = DailySequence(excludes=excludes)
        elif args.weekly is not None:
            seq = WeeklySequence(args.weekly)
        elif args.monthly is not None:
            excludes = [parse_mmdd(elem) for elem in args.exclude]
            seq = MonthlySequence(args.monthly, excludes=excludes)
        elif args.yearly is not None:
            excludes = [parse_date(elem) for elem in args.exclude]
            seq = YearlySequence(args.yearly, excludes=excludes)
        elif args.preset is not None:
            seq = Sequence.from_resource(args.preset)
    except ValueError as e:
        parser.error(str(e))
    assert seq is not None, "Unsupported sequence?!"
    return seq


def _eval_escape(m: re.Match) -> str:
    simple = {
        "\\": "\\",
        "0": "\x00",
        "a": "\a",
        "b": "\b",
        "f": "\f",
        "n": "\n",
        "r": "\r",
        "t": "\t",
        "v": "\v",
    }
    val = m.group(1)
    if len(val) == 1:
        return simple[val]
    if val[0] in "xX":
        return chr(int(val[1:], 16))
    return chr(int(val, 8))


def escaped_str(arg: str) -> str:
    return re.sub(
        r"\\([\\0abfnrtv]|x[0-9a-f]{2}|[0-3][0-7]{2})", _eval_escape, arg, flags=re.I
    )


SEP_EPILOG = """
SEPARATORS: separators can be any string of characters, with some escape
sequences evaluated:
* \\0, \\a, \\b, \\f, \\n, \\r, \\t, \\v: NUL, BEL, BS, FF, LF, CR, TAB, VT
* \\xhh: character with hex value hh
* \\ooo: character with octal value ooo
* \\\\: literal \\
"""


def add_sep_arg(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--sep",
        type=escaped_str,
        default="\n",
        help="output separator, see SEPARATORS for special values",
    )



import os.path
from abc import ABC, abstractmethod
from datetime import date, timedelta
from typing import Container, Dict, Iterable, Iterator, Optional, Tuple, Type, Union

from .calendar import (
    MonthInYear,
    Weekday,
    day_exists,
    parse_date,
    parse_mmdd,
    to_weekday,
)
from .data import load_yaml


class Sequence(ABC):
    """Abstract representation of a sequence of dates

    Minimal implementation requirement: ``__contains__``. Implementing ``next``
    and ``previous`` is highly recommended for efficiency.
    """

    @abstractmethod
    def __contains__(self, reference: date) -> bool:
        raise NotImplementedError

    def next(self, reference: date, strict: bool = True) -> date:
        """Return the next date in the sequence, after ``reference``.
        If ``strict`` is ``False`` and ``reference`` is in the sequence, it is
        returned.
        """
        if not strict and reference in self:
            return reference
        oneday = timedelta(days=1)
        current = reference + oneday
        while current not in self:
            current += oneday
        return current

    def previous(self, reference: date, strict: bool = True) -> date:
        """Return the previous date in the sequence, before ``reference``.
        If ``strict`` is ``False`` and ``reference`` is in the sequence, it is
        returned.
        """
        if not strict and reference in self:
            return reference
        oneday = timedelta(days=1)
        current = reference - oneday
        while current not in self:
            current -= oneday
        return current

    def nearest(self, reference: date, resolve: str = "previous") -> date:
        """Return the date closest to ``reference`` in the sequence.
        In case this is ambiguous, ``resolve`` defines which date to use
        (``"previous"`` or ``"next"``).
        """
        if resolve not in ["previous", "next"]:
            raise ValueError('`resolve` must be either "previous" or "next"')
        before = self.previous(reference, strict=False)
        after = self.next(reference, strict=False)
        delta_b = reference - before
        delta_a = after - reference
        if delta_b < delta_a:
            return before
        elif delta_b > delta_a:
            return after
        elif resolve == "previous":
            return before
        else:
            return after

    def range(
        self,
        start: date,
        end: date,
        include_start: bool = True,
        include_end: bool = True,
    ) -> Iterator[date]:
        """Return all matching dates between ``start`` and ``end``.
        ``include_start`` and ``include_end`` control whether ``start`` and
        ``end`` are returned if they are in the sequence.
        """
        if end < start:
            raise ValueError("Start date should be before end date")

        current = self.next(start, strict=(not include_start))
        last = self.previous(end, strict=(not include_end))
        while current <= last:
            yield current
            current = self.next(current)

    def bracket(
        self, reference: date, num: Union[int, Tuple[int, int]] = 1, strict: bool = True
    ) -> Iterator[date]:
        """Return matching dates around ``reference``

        Parameters
        ----------
        reference: :class:`datetime.date`
            Reference date
        num: int or (int, int) tuple
            Number of dates to include either side of ``reference``. If a single
            value is given, it is used on both sides
        strict: bool
            If ``False`` and ``reference`` is in the sequence, it will be
            returned as well (not counted in ``num``)
        """
        if isinstance(num, int):
            before = after = num
        else:
            before, after = num

        if before <= 0 or after <= 0:
            raise ValueError("`num` values must be positive")

        start = reference
        for _ in range(before):
            start = self.previous(start)

        end = reference
        for _ in range(after):
            end = self.next(end)

        for current in self.range(start, end):
            if strict and current == reference:
                continue
            yield current

    @classmethod
    @abstractmethod
    def _from_dict(cls, seq_dict: dict) -> "Sequence":
        """Create a specific sequence from the given dictionary

        Dictionary contents can vary depending on the sequence. Frequent items are:
        * ``days``: list of recurring days
        * ``excludes``: specification of which days to skip
        """
        raise NotImplementedError

    _known_types: Dict[str, Type["Sequence"]] = {}

    @classmethod
    def _register_sequence(cls, name: str, class_: Type["Sequence"]):
        assert name not in cls._known_types
        cls._known_types[name] = class_

    @classmethod
    def __init_subclass__(cls, /, seqname: Optional[str] = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if seqname is not None:
            Sequence._register_sequence(seqname, cls)

    @classmethod
    def from_dict(cls, seq_dict: dict) -> "Sequence":
        """Create a sequence from the given dictionary

        The type of sequence is specified by the ``type`` key, and must match
        one of the known sequences, e.g. ``daily``, ``weekly``, ``monthly``,
        ``yearly``.

        Dictionary contents can vary depending on the sequence. Frequent items are:

        * ``days``: list of recurring days
        * ``excludes``: specification of which days to skip

        Raises :class:`ValueError` if the type is unknown
        """
        if "type" not in seq_dict:
            raise ValueError("Sequence dictionary must contain `type` key")
        type_ = seq_dict["type"]
        if type_ not in cls._known_types:
            raise ValueError(f"Unknown type {type_!r}")
        return cls._known_types[type_]._from_dict(seq_dict)

    @classmethod
    def from_resource(cls, name: str) -> "Sequence":
        """Load a sequence from a resource file

        ``name`` should be either the name of a known sequence (in
        ``earthkit.time.data.sequences`` or ``EARTHKIT_TIME_SEQ_PATH``,
        without the extension), or the path to a YAML file

        Raises :class:`FileNotFoundError` if no corresponding resource is found
        """
        path = name if os.path.isfile(name) else None
        seq_dict = load_yaml(
            f"sequences/{name}.yaml", path, env_path="EARTHKIT_TIME_SEQ_PATH"
        )
        if not isinstance(seq_dict, dict):
            raise ValueError("Invalid resource file")
        return cls.from_dict(seq_dict)


class DailySequence(Sequence, seqname="daily"):
    """Sequence of consecutive dates

    Any day number (in the month) present in ``excludes`` will be skipped

    Can be created from a :class:`dict` with items:

    * ``type``: ``"daily"``
    * ``excludes``: (list of int, optional) days of the month to exclude
    """

    def __init__(self, excludes: Container[int] = set()):
        self.excludes = excludes

    def __contains__(self, reference: date) -> bool:
        return reference.day not in self.excludes

    def __repr__(self) -> str:
        return f"DailySequence(excludes={self.excludes!r})"

    @classmethod
    def _from_dict(cls, seq_dict: dict) -> Sequence:
        return cls(excludes=set(seq_dict.get("excludes", set())))


class WeeklySequence(Sequence, seqname="weekly"):
    """Sequence of dates happening on given days of each week

    Can be created from a :class:`dict` with items:

    * ``type``: ``"weekly"``
    * ``days``: (int, str, list of int, list of str) days of the week, either
      numeric (0 = Monday, ..., 6 = Sunday) or unambiguous prefixes of names
      (e.g. ``"M"``, ``"tue"``, ``"Friday"``)
    """

    def __init__(self, days: Union[int, Weekday, Iterable[int], Iterable[Weekday]]):
        if isinstance(days, (int, Weekday)):
            self.days = [Weekday(days)]
        else:
            self.days = sorted(Weekday(day) for day in days)
        if not self.days:
            raise ValueError("`days` cannot be empty")

    def __contains__(self, reference: date) -> bool:
        wday = Weekday(reference.weekday())
        return wday in self.days

    def __repr__(self) -> str:
        repr_days = ", ".join([day.name for day in self.days])
        return f"WeeklySequence(days=[{repr_days}])"

    def next(self, reference: date, strict: bool = True) -> date:
        wday = Weekday(reference.weekday())
        if wday in self.days and not strict:
            return reference
        new_wday = next((day for day in self.days if day > wday), self.days[0])
        delta = new_wday - wday
        assert -7 < delta < 7
        if delta > 0:
            return reference + timedelta(days=delta)
        else:
            return reference + timedelta(days=(delta + 7))

    def previous(self, reference: date, strict: bool = True) -> date:
        wday = Weekday(reference.weekday())
        if wday in self.days and not strict:
            return reference
        new_wday = next((day for day in self.days[::-1] if day < wday), self.days[-1])
        delta = new_wday - wday
        assert -7 < delta < 7
        if delta < 0:
            return reference + timedelta(days=delta)
        else:
            return reference + timedelta(days=(delta - 7))

    @classmethod
    def _from_dict(cls, seq_dict: dict) -> Sequence:
        if "days" not in seq_dict:
            raise ValueError("Weekly sequence must provide `days`")
        days = seq_dict["days"]
        if isinstance(days, (int, str)):
            days = [to_weekday(days)]
        else:
            days = [to_weekday(day) for day in days]
        return cls(days)


class MonthlySequence(Sequence, seqname="monthly"):
    """Sequence of dates happening on given days of each month

    Any ``(month, day)`` tuple present in ``excludes`` will be skipped

    Can be created from a :class:`dict` with items:

    * ``type``: ``"monthly"``
    * ``days``: (int, list of int) days of the month (1-31)
    * ``excludes``: (list of pairs of int, list of str, optional) days of the
      year to exclude, either in (month, day) or in "MMDD" form
    """

    def __init__(
        self,
        days: Union[int, Iterable[int]],
        excludes: Container[Tuple[int, int]] = set(),
    ):
        if isinstance(days, int):
            self.days = [days]
        else:
            self.days = sorted(days)
        if not self.days:
            raise ValueError("`days` cannot be empty")
        if any((day < 1 or day > 31) for day in self.days):
            raise ValueError("All days must be between 1 and 31")
        self.excludes = excludes

    def __contains__(self, reference: date) -> bool:
        return (
            reference.day in self.days
            and (reference.month, reference.day) not in self.excludes
        )

    def __repr__(self) -> str:
        return f"MonthlySequence(days={self.days!r}, excludes={self.excludes!r})"

    def next(self, reference: date, strict: bool = True) -> date:
        if not strict and reference in self:
            return reference
        ymonth = MonthInYear(reference.year, reference.month)
        new_day = next(
            (
                day
                for day in self.days
                if day > reference.day
                and day in ymonth
                and (ymonth.month, day) not in self.excludes
            ),
            None,
        )
        while new_day is None:
            ymonth = ymonth.next()
            for day in self.days:
                if day in ymonth and (ymonth.month, day) not in self.excludes:
                    new_day = day
                    break
        return date(ymonth.year, ymonth.month, new_day)

    def previous(self, reference: date, strict: bool = True) -> date:
        if not strict and reference in self:
            return reference
        ymonth = MonthInYear(reference.year, reference.month)
        new_day = next(
            (
                day
                for day in self.days[::-1]
                if day < reference.day
                and day in ymonth
                and (ymonth.month, day) not in self.excludes
            ),
            None,
        )
        while new_day is None:
            ymonth = ymonth.previous()
            for day in self.days[::-1]:
                if day in ymonth and (ymonth.month, day) not in self.excludes:
                    new_day = day
                    break
        return date(ymonth.year, ymonth.month, new_day)

    @classmethod
    def _from_dict(cls, seq_dict: dict) -> Sequence:
        if "days" not in seq_dict:
            raise ValueError("Monthly sequence must provide `days`")
        excludes = {parse_mmdd(exc) for exc in seq_dict.get("excludes", set())}
        return cls(seq_dict["days"], excludes=excludes)


class YearlySequence(Sequence, seqname="yearly"):
    """Sequence of dates happening on given days of each year (in (month, day) format)

    Can be created from a :class:`dict` with items:

    * ``type``: ``"yearly"``
    * ``days``: (str, pair of int, list of str, list of pairs of int) days of the year
      either in "MMDD" or in (month, day) form (1-12, 1-31)
    * ``excludes``: (list of str, list of triples of int) dates to exclude,
      either in "YYYYMMDD" or in (year, month, day) form
    """

    def __init__(
        self,
        days: Union[Tuple[int, int], Iterable[Tuple[int, int]]],
        excludes: Container[date] = set(),
    ):
        if (
            isinstance(days, tuple)
            and len(days) == 2
            and all(isinstance(day, int) for day in days)
        ):
            self.days = [days]
        else:
            self.days = sorted(days)
        self.excludes = excludes

    def __contains__(self, reference: date) -> bool:
        return (
            reference.month,
            reference.day,
        ) in self.days and reference not in self.excludes

    def __repr__(self) -> str:
        return f"YearlySequence(days={self.days!r}, excludes={self.excludes!r})"

    def next(self, reference: date, strict: bool = True) -> date:
        if not strict and reference in self:
            return reference

        year = reference.year
        new_month, new_day = next(
            (
                (month, day)
                for month, day in self.days
                if day_exists(year, month, day)
                and (month, day) > (reference.month, reference.day)
                and date(year, month, day) not in self.excludes
            ),
            (None, None),
        )
        while new_day is None:
            year += 1
            for month, day in self.days:
                if (
                    day_exists(year, month, day)
                    and date(year, month, day) not in self.excludes
                ):
                    new_month = month
                    new_day = day
                    break
        return date(year, new_month, new_day)

    def previous(self, reference: date, strict: bool = True) -> date:
        if not strict and reference in self:
            return reference

        year = reference.year
        new_month, new_day = next(
            (
                (month, day)
                for month, day in self.days[::-1]
                if day_exists(year, month, day)
                and (month, day) < (reference.month, reference.day)
                and date(year, month, day) not in self.excludes
            ),
            (None, None),
        )
        while new_day is None:
            year -= 1
            for month, day in self.days[::-1]:
                if (
                    day_exists(year, month, day)
                    and date(year, month, day) not in self.excludes
                ):
                    new_month = month
                    new_day = day
                    break
        return date(year, new_month, new_day)

    @classmethod
    def _from_dict(cls, seq_dict: dict) -> Sequence:
        if "days" not in seq_dict:
            raise ValueError("Yearly sequence must provide `days`")
        days = seq_dict["days"]
        if isinstance(days, str) or isinstance(days[0], int):
            days = [parse_mmdd(days)]
        else:
            days = [parse_mmdd(day) for day in days]
        excludes = {parse_date(exc) for exc in seq_dict.get("excludes", set())}
        return cls(days, excludes=excludes)



import argparse
from datetime import timedelta
from typing import List, Optional

from ..calendar import parse_date
from .actions import ActionParser
from .cliout import format_date


def date_shift_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    delta = timedelta(days=args.days)
    print(format_date(args.date + delta))


def date_diff_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    delta = args.date1 - args.date2
    print(delta.days)


def get_parser() -> argparse.ArgumentParser:
    parser = ActionParser(description="Date manipulation tools")

    shift_action = parser.add_action(
        "shift",
        date_shift_action,
        help="shift a date",
        description="Shift a date by the given number of days",
    )
    shift_action.add_argument("date", type=parse_date, help="reference date")
    shift_action.add_argument("days", type=int, help="number of days (can be negative)")

    diff_action = parser.add_action(
        "diff",
        date_diff_action,
        help="subtract two dates",
        description="Subtract DATE2 from DATE1, returning the number of days",
    )
    diff_action.add_argument("date1", type=parse_date, help="first date (+)")
    diff_action.add_argument("date2", type=parse_date, help="second date (-)")

    return parser


def main(sys_args: Optional[List[str]] = None):
    parser = get_parser()
    args = parser.parse_args(sys_args)
    if args.action is None:
        parser.error("Please specify an action")
    args.action(parser, args)


if __name__ == "__main__":
    main()



import os

import pytest

from earthkit.time.data import ResourceType, find_resource


def test_find_resource(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
):
    # Packaged resources only
    assert find_resource("sequences/ecmwf-mon-thu.yaml")[0] == ResourceType.PACKAGED
    assert find_resource("sequences/nonexistent")[0] == ResourceType.NOTFOUND

    custom = tmp_path_factory.mktemp("custom")
    custom_foo = custom / "foo.txt"
    custom_foo.write_text("Hello!")
    custom_seq = custom / "ecmwf-mon-thu.yaml"
    custom_seq.write_text("type: weekly\ndays: [0, 3]\n")

    # Resource by path
    assert find_resource("sequences/test-hello", path=str(custom_foo)) == (
        ResourceType.FILE,
        str(custom_foo),
    )

    # Path overrides packaged
    assert find_resource("sequences/ecmwf-mon-thu.yaml", path=str(custom_foo)) == (
        ResourceType.FILE,
        str(custom_foo),
    )

    # Non-existent path does not override
    assert (
        find_resource(
            "sequences/ecmwf-mon-thu.yaml", path=str(custom / "nonexistent.txt")
        )[0]
        == ResourceType.PACKAGED
    )

    # Resource by env file
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(custom_foo))
        assert find_resource("sequences/test-hello", env_file="TEST_RES_FILE") == (
            ResourceType.FILE,
            str(custom_foo),
        )

    # Non-existent env file
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(custom / "nonexistent.txt"))
        assert (
            find_resource("sequences/test-hello", env_file="TEST_RES_FILE")[0]
            == ResourceType.NOTFOUND
        )

    # Env file overrides packaged
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(custom_seq))
        assert find_resource(
            "sequences/ecmwf-mon-thu.yaml", env_file="TEST_RES_FILE"
        ) == (
            ResourceType.FILE,
            str(custom_seq),
        )

    # Non-existent env file does not override
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(custom / "nonexistent.txt"))
        assert (
            find_resource("sequences/ecmwf-mon-thu.yaml", env_file="TEST_RES_FILE")[0]
            == ResourceType.PACKAGED
        )

    other = tmp_path_factory.mktemp("other")
    other_foo = other / "foo.txt"
    other_foo.write_text("Hi there!")
    other_bar = other / "bar.txt"
    other_bar.write_text("Testing")

    # Path overrides env file
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(custom_foo))
        assert find_resource(
            "sequences/test-hello", path=str(other_foo), env_file="TEST_RES_FILE"
        ) == (
            ResourceType.FILE,
            str(other_foo),
        )

    # Resource by env path, with precedence
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(other), str(custom)]))
        assert find_resource("sequences/foo.txt", env_path="TEST_RES_DIR") == (
            ResourceType.FILE,
            str(other_foo),
        )

    # Resource by env path, not present in first dir
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(custom), str(other)]))
        assert find_resource("sequences/bar.txt", env_path="TEST_RES_DIR") == (
            ResourceType.FILE,
            str(other_bar),
        )

    # Resource by env path, non-existent
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(custom), str(other)]))
        assert (
            find_resource("sequences/nonexistent", env_path="TEST_RES_DIR")[0]
            == ResourceType.NOTFOUND
        )

    # Env path overrides packaged
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(other), str(custom)]))
        assert find_resource(
            "sequences/ecmwf-mon-thu.yaml", env_path="TEST_RES_DIR"
        ) == (
            ResourceType.FILE,
            str(custom_seq),
        )

    # Non-existent env path does not override
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(other / "nonexistent")]))
        assert (
            find_resource("sequences/ecmwf-mon-thu.yaml", env_path="TEST_RES_DIR")[0]
            == ResourceType.PACKAGED
        )

    # Env file overrides env path
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_FILE", str(other_bar))
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(other), str(custom)]))
        assert find_resource(
            "sequences/foo.txt", env_file="TEST_RES_FILE", env_path="TEST_RES_DIR"
        ) == (
            ResourceType.FILE,
            str(other_bar),
        )

    # Path overrides env path
    with monkeypatch.context() as m:
        m.setenv("TEST_RES_DIR", os.pathsep.join([str(other), str(custom)]))
        assert find_resource(
            "sequences/foo.txt", path=str(other_bar), env_path="TEST_RES_DIR"
        ) == (
            ResourceType.FILE,
            str(other_bar),
        )



import argparse
from typing import List, Optional

from ..calendar import parse_date
from .actions import ActionParser
from .cliargs import (
    SEP_EPILOG,
    SEQ_EPILOG,
    add_sep_arg,
    add_sequence_args,
    create_sequence,
)
from .cliout import format_date, format_date_list


def seq_next_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    seq = create_sequence(parser, args)
    new = seq.next(args.date, strict=(not args.inclusive))
    for _ in range(args.skip):
        new = seq.next(new, strict=True)
    print(format_date(new))


def seq_prev_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    seq = create_sequence(parser, args)
    new = seq.previous(args.date, strict=(not args.inclusive))
    for _ in range(args.skip):
        new = seq.previous(new, strict=True)
    print(format_date(new))


def seq_nearest_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    seq = create_sequence(parser, args)
    print(format_date(seq.nearest(args.date, resolve=args.resolve)))


def seq_range_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    seq = create_sequence(parser, args)
    print(
        format_date_list(
            seq.range(
                getattr(args, "from"),
                args.to,
                (not args.exclude_start),
                (not args.exclude_end),
            ),
            sep=args.sep,
        )
    )


def seq_bracket_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    num = args.before
    if args.after is not None:
        num = (args.before, args.after)
    seq = create_sequence(parser, args)
    print(
        format_date_list(
            seq.bracket(args.date, num, strict=(not args.inclusive)), sep=args.sep
        )
    )


def get_parser() -> argparse.ArgumentParser:
    parser = ActionParser(
        description="Manipulate sequences of dates", fromfile_prefix_chars="@"
    )

    next_action = parser.add_action(
        "next",
        seq_next_action,
        help="compute the next date in the given sequence",
        description="Compute the next date in the given sequence",
        epilog=SEQ_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    next_action.add_argument("date", type=parse_date, help="reference date")
    add_sequence_args(next_action)
    next_action.add_argument(
        "--inclusive",
        action="store_true",
        help="if the given date is in the sequence, return it",
    )
    next_action.add_argument(
        "--skip",
        type=int,
        default=0,
        help="if set, skip over that number of dates",
    )

    prev_action = parser.add_action(
        "previous",
        seq_prev_action,
        help="compute the previous date in the given sequence",
        description="Compute the previous date in the given sequence",
        epilog=SEQ_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    prev_action.add_argument("date", type=parse_date, help="reference date")
    add_sequence_args(prev_action)
    prev_action.add_argument(
        "--inclusive",
        action="store_true",
        help="if the given date is in the sequence, return it",
    )
    prev_action.add_argument(
        "--skip",
        type=int,
        default=0,
        help="if set, skip over that number of dates",
    )

    nearest_action = parser.add_action(
        "nearest",
        seq_nearest_action,
        help="compute the nearest date in the given sequence",
        description="Compute the nearest date in the given sequence",
        epilog=SEQ_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    nearest_action.add_argument("date", type=parse_date, help="reference date")
    add_sequence_args(nearest_action)
    nearest_action.add_argument(
        "--resolve",
        choices=("previous", "next"),
        default="previous",
        help="return this date in case of a tie",
    )

    range_action = parser.add_action(
        "range",
        seq_range_action,
        help="compute the sequence dates that fall within a range",
        description="Compute the sequence dates that fall winthin the given range",
        epilog=SEQ_EPILOG + "\n" + SEP_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    add_sequence_args(range_action)
    add_sep_arg(range_action)
    range_action.add_argument("from", type=parse_date, help="starting date")
    range_action.add_argument("to", type=parse_date, help="ending date")
    range_action.add_argument(
        "--exclude-start", action="store_true", help="exclude starting date"
    )
    range_action.add_argument(
        "--exclude-end", action="store_true", help="exclude ending date"
    )

    bracket_action = parser.add_action(
        "bracket",
        seq_bracket_action,
        help="compute the sequence dates around a date",
        description="Compute the sequence dates around the given date",
        epilog=SEQ_EPILOG + "\n" + SEP_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    add_sequence_args(bracket_action)
    add_sep_arg(bracket_action)
    bracket_action.add_argument("date", type=parse_date, help="reference date")
    bracket_action.add_argument(
        "before",
        nargs="?",
        type=int,
        default=1,
        help="number of dates before the given date (default 1)",
    )
    bracket_action.add_argument(
        "after",
        nargs="?",
        type=int,
        default=None,
        help="number of dates after the given date (default: same as before)",
    )
    bracket_action.add_argument(
        "--inclusive",
        action="store_true",
        help="include the given date in the sequence",
    )

    return parser


def main(sys_args: Optional[List[str]] = None):
    parser = get_parser()
    args = parser.parse_args(sys_args)
    if args.action is None:
        parser.error("Please specify an action")
    args.action(parser, args)


if __name__ == "__main__":
    main()



import argparse
import textwrap
from typing import List, Optional

from ..calendar import parse_date
from ..climatology import date_range, model_climate_dates
from .actions import ActionParser
from .cliargs import (
    SEP_EPILOG,
    SEQ_EPILOG,
    add_sep_arg,
    add_sequence_args,
    create_sequence,
)
from .cliout import format_date_list


def date_range_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    dates = date_range(args.date, args.start, args.end)
    print(format_date_list(dates, sep=args.sep))


def model_climate_action(parser: argparse.ArgumentParser, args: argparse.Namespace):
    reference = args.date
    start = args.start
    end = args.end
    before = args.before
    after = args.after
    seq = create_sequence(parser, args)
    dates = model_climate_dates(reference, start, end, before, after, seq)
    print(format_date_list(dates, sep=args.sep))


def get_parser() -> argparse.ArgumentParser:
    parser = ActionParser(
        description="Tools to compute model climate dates", fromfile_prefix_chars="@"
    )

    range_action = parser.add_action(
        "range",
        date_range_action,
        help="compute climatological date ranges",
        description="Compute climatological date ranges, one day per year in a given range",
        epilog=SEP_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    range_action.add_argument("date", type=parse_date, help="reference date")

    range_start_group = range_action.add_mutually_exclusive_group(required=True)
    range_start_group.add_argument(
        "--from-date", type=parse_date, dest="start", help="starting date"
    )
    range_start_group.add_argument(
        "--from-year", type=int, dest="start", help="starting year"
    )

    range_end_group = range_action.add_mutually_exclusive_group(required=True)
    range_end_group.add_argument(
        "--to-date", type=parse_date, dest="end", help="ending date"
    )
    range_end_group.add_argument("--to-year", type=int, dest="end", help="ending year")

    add_sep_arg(range_action)

    mclim_action = parser.add_action(
        "mclim",
        model_climate_action,
        help="compute sets of dates for model climatologies",
        description=textwrap.dedent(
            """\
            Compute sets of dates for model climatologies

            This combines a climatological range (same day in multiple years) and a
            recurring source (e.g. twice a week).
            """
        ),
        epilog=SEQ_EPILOG + "\n" + SEP_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    mclim_action.add_argument("date", type=parse_date, help="reference date")

    mclim_start_group = mclim_action.add_mutually_exclusive_group(required=True)
    mclim_start_group.add_argument(
        "--from-date", type=parse_date, dest="start", help="starting date"
    )
    mclim_start_group.add_argument(
        "--from-year", type=int, dest="start", help="starting year"
    )

    mclim_end_group = mclim_action.add_mutually_exclusive_group(required=True)
    mclim_end_group.add_argument(
        "--to-date", type=parse_date, dest="end", help="ending date"
    )
    mclim_end_group.add_argument("--to-year", type=int, dest="end", help="ending year")

    mclim_action.add_argument(
        "--before",
        required=True,
        metavar="NUM",
        type=int,
        help="pick up all inputs starting from %(metavar)s days before the chosen date",
    )
    mclim_action.add_argument(
        "--after",
        required=True,
        metavar="NUM",
        type=int,
        help="pick up all inputs up to %(metavar)s days after the chosen date",
    )

    add_sequence_args(mclim_action)
    add_sep_arg(mclim_action)

    return parser


def main(sys_args: Optional[List[str]] = None):
    parser = get_parser()
    args = parser.parse_args(sys_args)
    if args.action is None:
        parser.error("Please specify an action")
    args.action(parser, args)


if __name__ == "__main__":
    main()



from datetime import date
from typing import Iterable


def format_date(d: date) -> str:
    return d.strftime("%Y%m%d")


def format_date_list(dates: Iterable[date], sep: str = "/") -> str:
    return sep.join(format_date(d) for d in dates)



from .climatology import date_range, model_climate_dates
from .sequence import (
    DailySequence,
    MonthlySequence,
    Sequence,
    WeeklySequence,
    YearlySequence,
)

__all__ = [
    "date_range",
    "model_climate_dates",
    "DailySequence",
    "MonthlySequence",
    "Sequence",
    "WeeklySequence",
    "YearlySequence",
]



from contextlib import nullcontext
from datetime import date
from typing import Tuple, Union

import pytest

from earthkit.time.calendar import (
    MonthInYear,
    Weekday,
    day_exists,
    month_length,
    parse_date,
    parse_mmdd,
    to_weekday,
)


@pytest.mark.parametrize(
    "arg, expected",
    [
        (-1, "^Week day out of range: "),
        (0, Weekday.MONDAY),
        (1, Weekday.TUESDAY),
        (7, "^Week day out of range: "),
        ("3", Weekday.THURSDAY),
        ("5", Weekday.SATURDAY),
        ("8", "^Week day out of range: "),
        ("M", Weekday.MONDAY),
        ("T", "^Ambiguous week day: "),
        ("wed", Weekday.WEDNESDAY),
        ("Fri", Weekday.FRIDAY),
        ("SUNDAY", Weekday.SUNDAY),
        ("Notaday", "^Unrecognised week day: "),
    ],
)
def test_to_weekday(arg: Union[int, str], expected: Union[Weekday, str]):
    context = nullcontext()
    if isinstance(expected, str):
        context = pytest.raises(ValueError, match=expected)
    with context:
        assert to_weekday(arg) == expected


@pytest.mark.parametrize(
    "year, month, expected",
    [
        (2000, 1, 31),
        (2001, 2, 28),
        (2002, 3, 31),
        (2003, 4, 30),
        (2004, 5, 31),
        (2005, 6, 30),
        (2006, 7, 31),
        (2007, 8, 31),
        (2008, 9, 30),
        (2009, 10, 31),
        (2010, 11, 30),
        (2011, 12, 31),
        (2012, 2, 29),
        (2013, 13, "^Invalid month: 13$"),
    ],
)
def test_month_length(year: int, month: int, expected: Union[int, str]):
    context = nullcontext()
    if isinstance(expected, str):
        context = pytest.raises(ValueError, match=expected)
    with context:
        assert month_length(year, month) == expected


@pytest.mark.parametrize(
    "year, month, day, expected",
    [
        (1999, 0, 12, False),
        (2000, 1, 32, False),
        (2001, 2, 29, False),
        (2002, 13, 18, False),
        (2003, 8, -5, False),
        (2004, 2, 29, True),
        (2005, 6, 31, False),
        (2006, 7, 31, True),
    ],
)
def test_day_exists(year: int, month: int, day: int, expected: bool):
    assert day_exists(year, month, day) == expected


@pytest.mark.parametrize(
    "year, month, ok", [(y, y - 2009, y > 2009 and y < 2022) for y in range(2005, 2025)]
)
def test_monthinyear_create(year: int, month: int, ok: bool):
    context = (
        nullcontext() if ok else pytest.raises(ValueError, match="^Invalid month:")
    )
    with context:
        ymonth = MonthInYear(year, month)
    if ok:
        assert ymonth.year == year
        assert ymonth.month == month


@pytest.mark.parametrize(
    "year, month, day, expected",
    [
        (2010, 1, 3, True),
        (2011, 2, 29, False),
        (2012, 2, 29, True),
        (2013, 3, date(2012, 3, 1), False),
        (2014, 4, date(2014, 3, 8), False),
        (2015, 5, date(2015, 5, 28), True),
        (2016, 6, 31, False),
        (2017, 7, 32, False),
        (2018, 8, 0, False),
    ],
)
def test_monthinyear_contains(
    year: int, month: int, day: Union[int, date], expected: bool
):
    ymonth = MonthInYear(year, month)
    assert (day in ymonth) == expected


@pytest.mark.parametrize(
    "year, month, expected",
    [
        (1993, 1, 31),
        (1994, 2, 28),
        (1996, 2, 29),
        (1995, 3, 31),
        (1997, 4, 30),
        (1998, 5, 31),
        (1999, 6, 30),
        (2000, 7, 31),
        (2001, 8, 31),
        (2002, 9, 30),
        (2003, 10, 31),
        (2004, 11, 30),
        (2005, 12, 31),
    ],
)
def test_monthinyear_length(year: int, month: int, expected: int):
    ymonth = MonthInYear(year, month)
    assert ymonth.length() == expected


@pytest.mark.parametrize(
    "year, month, eyear, emonth",
    [(1954, 3, 1954, 4), (2020, 12, 2021, 1), (2005, 1, 2005, 2), (1976, 11, 1976, 12)],
)
def test_monthinyear_next(year: int, month: int, eyear: int, emonth: int):
    ymonth = MonthInYear(year, month)
    next_ymonth = ymonth.next()
    assert next_ymonth.year == eyear
    assert next_ymonth.month == emonth


@pytest.mark.parametrize(
    "year, month, eyear, emonth",
    [(2050, 6, 2050, 5), (1972, 1, 1971, 12), (1999, 2, 1999, 1), (2013, 12, 2013, 11)],
)
def test_monthinyear_previous(year: int, month: int, eyear: int, emonth: int):
    ymonth = MonthInYear(year, month)
    prev_ymonth = ymonth.previous()
    assert prev_ymonth.year == eyear
    assert prev_ymonth.month == emonth


@pytest.mark.parametrize(
    "arg, expected",
    [
        ((5,), "^not enough values to unpack"),
        ((10, 3), (10, 3)),
        ((4, 31), "^Invalid day: "),
        ("14", "^Unrecognised month-day value: "),
        ("test", "^Unrecognised month-day value: "),
        ("1304", "^Invalid month: "),
        ("0230", "^Invalid day: "),
        ("0906", (9, 6)),
        ("1213", (12, 13)),
    ],
)
def test_parse_mmdd(
    arg: Union[Tuple[int, int], str], expected: Union[Tuple[int, int], str]
):
    context = nullcontext()
    if isinstance(expected, str):
        context = pytest.raises(ValueError, match=expected)
    with context:
        assert parse_mmdd(arg) == expected


@pytest.mark.parametrize(
    "arg, expected",
    [
        pytest.param("", None, id="empty"),
        pytest.param("2020", None, id="yearonly"),
        pytest.param("202005", None, id="yearmonthonly"),
        pytest.param("20202503", None, id="notadate"),
        pytest.param("20201204", (2020, 12, 4), id="ok"),
        pytest.param("202010251200", None, id="toolong"),
        pytest.param((2022,), "^not enough values to unpack", id="tup-inclomplete"),
        pytest.param((2022, 2, 3), (2022, 2, 3), id="tup-ok"),
        pytest.param((2022, 2, 30), "^Invalid date: ", id="tup-ok"),
    ],
)
def test_parse_date(arg: str, expected: Union[Tuple[int, int, int], None]):
    context = nullcontext()
    if expected is None:
        context = pytest.raises(ValueError, match="^Unrecognised date format: ")
    elif isinstance(expected, str):
        context = pytest.raises(ValueError, match=expected)
    else:
        expected = date(*expected)
    with context:
        assert parse_date(arg) == expected



import enum
import os
import pathlib
from importlib import resources
import types
from typing import Optional, TextIO, Tuple, Union

import yaml

# importlib.resources.Packages is removed from Python >= 3.13
Package = Union[types.ModuleType, str]

PACKAGE = "earthkit.time.data"


def _extract_package(path: str) -> Tuple[str, str]:
    res = pathlib.PurePosixPath(path)
    parts = list(res.parts)
    if len(parts) == 1:
        return PACKAGE, parts[0]
    return (PACKAGE + "." + ".".join(parts[:-1])), parts[-1]


def _is_resource(package: Package, name: str) -> bool:
    if hasattr(resources, "files"):  # Python >= 3.9
        return resources.files(package).joinpath(name).is_file()
    else:
        return resources.is_resource(package, name)


def _open_text(
    package: Package,
    resource: str,
    encoding: str = "utf-8",
    errors: str = "strict",
) -> TextIO:
    if hasattr(resources, "files"):  # Python >= 3.9
        return (
            resources.files(package)
            .joinpath(resource)
            .open("r", encoding=encoding, errors=errors)
        )
    else:
        return resources.open_text(package, resource, encoding, errors)


class ResourceType(enum.Enum):
    NOTFOUND = enum.auto()
    FILE = enum.auto()
    PACKAGED = enum.auto()


def find_resource(
    name: str,
    path: Optional[str] = None,
    env_file: Optional[str] = None,
    env_path: Optional[str] = None,
) -> Tuple[ResourceType, str]:
    """Find a given resource file using various methods

    The search order is: ``path``, ``env_file``, `env_path``, packaged date.

    Parameters
    ----------
    name: str
        Resource file name
    path: str or None
        Candidate path to the resource
    env_file: str or None
        Environment variable to check for a path to the file
    env_path: str or None
        Environment variable to check for a set of directories

    Returns
    -------
    ResourceType
        Whether the resource has been found, and where
    str
        Path to the resource, if found
    """

    if path is not None:
        if pathlib.Path(path).is_file():
            return ResourceType.FILE, path
    if env_file is not None:
        candidate = os.getenv(env_file)
        if candidate is not None:
            if pathlib.Path(candidate).is_file():
                return ResourceType.FILE, candidate
    if env_path is not None:
        path = os.getenv(env_path)
        bname = pathlib.Path(name).name
        if path is not None:
            for dir in path.split(os.pathsep):
                candidate = pathlib.Path(dir) / bname
                if candidate.is_file():
                    return ResourceType.FILE, str(candidate)
    pkg, res = _extract_package(name)
    if _is_resource(pkg, res):
        return ResourceType.PACKAGED, name
    return ResourceType.NOTFOUND, ""


def load_yaml(
    name: str,
    path: Optional[str] = None,
    env_file: Optional[str] = None,
    env_path: Optional[str] = None,
) -> object:
    """Load a YAML resource

    See `find_resource` for the signification of the parameters

    Raises `FileNotFoundError` if no corresponding resource is found
    """
    tp, res_path = find_resource(name, path, env_file, env_path)
    if tp == ResourceType.NOTFOUND:
        raise FileNotFoundError(name)
    elif tp == ResourceType.FILE:
        f = open(res_path, "r")
    elif tp == ResourceType.PACKAGED:
        pkg, res = _extract_package(res_path)
        f = _open_text(pkg, res)
    else:
        raise ValueError("Unknown resource type")
    with f:
        return yaml.safe_load(f)



from earthkit.time.utilities import merge_sorted


def test_merge_sorted():
    assert list(merge_sorted([])) == []
    assert list(merge_sorted([range(5)])) == list(range(5))
    assert list(merge_sorted([range(3), range(4, 7)])) == [0, 1, 2, 4, 5, 6]
    assert list(merge_sorted((range(n, 10, 2) for n in range(2)))) == list(range(10))
    assert list(merge_sorted((range(n, 10, 3) for n in range(3)))) == list(range(10))
    assert list(merge_sorted((range(n, 10, 4) for n in range(4)))) == list(range(10))



"""Date utilities to build a climatology"""

from datetime import date, timedelta
from typing import Iterator, Union

from .sequence import Sequence, YearlySequence
from .utilities import merge_sorted


def date_range(
    reference: date,
    start: Union[date, int],
    end: Union[date, int],
    recurrence: str = "yearly",
    include_endpoint: bool = True,
) -> Iterator[date]:
    """Generate a sequence of dates following a recurrence pattern

    If the reference date is February 29th on a leap year, it will be replaced
    by February 28th for every year in the output.

    Parameters
    ----------
    reference: :class:`datetime.date`
        Reference date setting the fixed part in the sequence (e.g., month and day
        for a yearly recurrence)
    start: :class:`datetime.date` or int
        Start of the period. Either a full date or a meaningful identifier (e.g.
        year for a yearly recurrence)
    end: :class:`datetime.date` or int
        End of the period. Included in the sequence unless ``include_endpoint`` is
        ``False``
    recurrence: "yearly"
        Type of recurrence
    include_endpoint: bool
        If ``False``, do not include the end in the sequence

    Returns
    -------
    :class:`datetime.date` iterator
        Sequence of dates

    Examples
    --------
    >>> list(date_range(date(2020, 4, 12), 1999, 2002))
    [datetime.date(1999, 4, 12), datetime.date(2000, 4, 12), datetime.date(2001, 4, 12), datetime.date(2002, 4, 12)]
    >>> list(date_range(date(2020, 4, 12), 1999, 2002, include_endpoint=False))
    [datetime.date(1999, 4, 12), datetime.date(2000, 4, 12), datetime.date(2001, 4, 12)]
    >>> list(date_range(date(2014, 8, 23), date(2010, 8, 16), date(2012, 8, 1)))
    [datetime.date(2010, 8, 23), datetime.date(2011, 8, 23)]
    """

    _known_recurrences = ["yearly"]
    if recurrence not in _known_recurrences:
        known = ", ".join(_known_recurrences)
        raise ValueError(f"Unknown recurrence {recurrence!r}, expected one of: {known}")

    if recurrence == "yearly":
        if reference.month == 2 and reference.day == 29:
            reference = reference.replace(day=28)

        if not isinstance(start, date):
            start = reference.replace(year=start)

        if not isinstance(end, date):
            end = reference.replace(year=end)

        seq = YearlySequence((reference.month, reference.day))
        yield from seq.range(start, end, include_end=include_endpoint)


def model_climate_dates(
    reference: date,
    start: Union[date, int],
    end: Union[date, int],
    before: Union[timedelta, int],
    after: Union[timedelta, int],
    sequence: Sequence,
) -> Iterator[date]:
    """Generate a set of dates for a model climate

    The set is created by combining yearly dates between ``start`` and ``end``,
    for each date between ``reference - before`` and ``reference + after``. If
    any of these dates is February 29th, the whole corresponing sequence will
    use February 28th instead.

    Parameters
    ----------
    reference: :class:`datetime.date`
        Reference date for the climate
    start: :class:`datetime.date` or int
        Start of the climatological period. Either a full date or a year
    end: :class:`datetime.date` or int
        End of the climatological period. Either a full date or a year
    before: :class:`datetime.timedelta` or int
        Cut-off before the reference date. Either a timedelta or a number of
        days
    after: :class:`datetime.timedelta` or int
        Cut-off after the reference date. Either a timedelta or a number of days
    sequence: :class:`earthkit.time.sequence.Sequence`
        Sequence of available dates in the reference set

    Returns
    -------
    :class:`datetime.date` iterator
        Sequence of dates

    Examples
    --------
    >>> from earthkit.time.calendar import MONDAY, THURSDAY
    >>> from earthkit.time import MonthlySequence, WeeklySequence
    >>> sequence = WeeklySequence([MONDAY, THURSDAY])
    >>> [f"{d:%Y%m%d}" for d in model_climate_dates(date(2024, 2, 12), 2020, 2023, 7, 7, sequence)]
    ... # doctest: +NORMALIZE_WHITESPACE
    ['20200205', '20200208', '20200212', '20200215', '20200219',
     '20210205', '20210208', '20210212', '20210215', '20210219',
     '20220205', '20220208', '20220212', '20220215', '20220219',
     '20230205', '20230208', '20230212', '20230215', '20230219']
    >>> sequence = MonthlySequence(range(1, 32, 2), excludes=[(2, 29)])
    >>> [f"{d:%Y%m%d}" for d in model_climate_dates(date(2024, 2, 12), 2020, 2023, 7, 7, sequence)]
    ... # doctest: +NORMALIZE_WHITESPACE
    ['20200205', '20200207', '20200209', '20200211', '20200213', '20200215', '20200217', '20200219',
     '20210205', '20210207', '20210209', '20210211', '20210213', '20210215', '20210217', '20210219',
     '20220205', '20220207', '20220209', '20220211', '20220213', '20220215', '20220217', '20220219',
     '20230205', '20230207', '20230209', '20230211', '20230213', '20230215', '20230217', '20230219']
    """
    if not isinstance(before, timedelta):
        before = timedelta(days=before)
    if not isinstance(after, timedelta):
        after = timedelta(days=after)
    yield from merge_sorted(
        date_range(d, start, end)
        for d in sequence.range(reference - before, reference + after)
    )
